
require 'test/unit'
require 'test/unit/testsuite'
require 'test/unit/ui/console/testrunner'
require 'amrita2'
require 'rexml/document'

# ruby -I~/amrita2/lib ~/amrita2/test/test_core.rb   |ruby -pe '$_.sub!(/^ */, "") ' 

module Amrita2Test
  include Amrita2
  include Amrita2::Core

  class TestSanitize < Test::Unit::TestCase
    include Amrita2

    def test_sanitize
      t = "<>&"
      assert_equal("&lt;&gt;&amp;", t.amrita_sanitize)
      assert_equal("1", 1.amrita_sanitize)
    end

    def test_simple
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa)
      end
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('&<>')
      end

      assert_equal("<x id='aaa'>&amp;&lt;&gt;</x>", out)
      out = ""
      t.expand_template(out) do |m|
        m.aaa(SanitizedString['</x> INJECTION CODE <x>'])
      end

      assert_equal("<x id='aaa'></x> INJECTION CODE <x></x>", out)
    end

    def test_use_block
      spec = Amrita2::define_template_spec do 
        dynamic_element(:link, :use_original_element=>true)
      end
      doc = REXML::Document.new  "<p>See <a id='link' href='http://www.ruby-lang.org/'>Ruby</a> for detail</p>"
      t = spec.create_template(doc)
      
      t.expand_template(out="") do |m|
        m.link('Ruby homepage<>&')
      end
      assert_equal("<p>See <a href='http://www.ruby-lang.org/' id='link'>Ruby homepage&lt;&gt;&amp;</a> for detail</p>", out)
    end

    def test_replace_attr
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :replace_attr=>true)
      end

      doc = REXML::Document.new  "<x id='aaa' class='c'>111</x>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx', :class=>'<>"\'&')
      end
      assert_equal("<x class='&lt;&gt;&quot;&apos;&amp;' id='aaa'>xxx</x>", out)
    end
  end


  class TestCodeGenerator < Test::Unit::TestCase
    include Amrita2

    def test_put_constant
      cg = CodeGenerator.new("out")
      cg.put_constant("aaa")
      cg.put_constant("bbb")
      cg.flush()
      assert_equal('out << "aaabbb"', cg.results.join)

      cg = CodeGenerator.new("out")
      cg.put_constant("aaa")
      cg.put_constant("bbb")
      assert_equal('out << "aaabbb"', cg.results.join)

      out = "111"
      eval cg.results.join
      assert_equal('111aaabbb', out)
    end

    def test_if
      cg = CodeGenerator.new("out")
      cg.if_('a==b') do
        cg.put_constant("xyz")
      end
      assert_equal('if a==b;  out << "xyz";end', cg.results.join(';'))

      a=b=1
      out = ""
      eval cg.results.join("\n")
      assert_equal('xyz',out )

      a=2
      out = ""
      eval cg.results.join("\n")
      assert_equal('',out )
    end

    def test_if_else
      cg = CodeGenerator.new("out")
      cg.if_('a==b') do
        cg.put_constant("xyz")
        cg.else_ do
          cg.put_constant("abc")
        end
      end

      #put_constant cg.results.join("\n")
      assert_equal("if a==b;  out << \"xyz\";else;  out << \"abc\";end", cg.results.join(';'))

      a=b=1
      out = ""
      eval cg.results.join("\n")
      assert_equal('xyz',out )

      a=2
      out = ""
      eval cg.results.join("\n")
      assert_equal('abc',out )
    end
  end

  class TestElementMatcher < Test::Unit::TestCase
    include Amrita2
    TEMPLATE_TEXT_EM = <<-END
    <html>
    <body>
      <h1 id="title">title</h1>
      <p id="text">text</p>
      <table>
        <tr><th></th><th></th></tr>
        <tr><td></td><td></td></tr>
      </table>
    </body>
    </html>
    END

    def setup
      @doc = REXML::Document.new TEMPLATE_TEXT_EM
      @root = @doc.root
    end

    def test_id_match
      m = IDMatcher.new(:title)
      e = @root.elements["body/h1"]
      assert(m.match(e))
      assert_equal(e, m.find(@root))

      m = IDMatcher.new(:text)
      e = @root.elements["body/p"]
      assert(m.match(e))
      assert_equal(e, m.find(@root))

      assert_nil(IDMatcher.new(:xxx).find(@root))
    end

    def test_tag_match
      m = TagMatcher.new(:table)
      e = @root.elements["body/table"]
      assert(m.match(e))
      assert_equal(e, m.find(@root))

      assert_equal(@root.elements["body/table/tr/td"], TagMatcher.new(:td).find(@root))
    end
  end

  class TestRuntime < Test::Unit::TestCase
    include Amrita2::Runtime

    def test_element
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      e = doc.root
      assert_equal("x", e.name)
      assert_equal("id", e.attributes.keys.first)
      assert_equal("aaa", e.attributes["id"])
      assert_equal("111", e.text)
      assert_equal("<x id='aaa'>111</x>", e.to_s)

      e = new_element(:x, :id=>"aaa")
      assert_equal("x", e.name)
      assert_equal("id", e.attributes.keys.first)
      assert_equal("aaa", e.attributes["id"])
      assert_equal(nil, e.text)
      assert_equal("<x id='aaa'/>", e.to_s)

      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      e = doc.root
      assert_equal("<x id='aaa'>", start_tag(e))
      assert_equal("</x>", end_tag(e))

      e = new_element(:x, :id=>"aaa", :xxx=>"yy'")
      e.delete_attribute("id")
      assert_equal("<x xxx='yy&apos;'>", start_tag(e))

    end

    def test_mainstream
      out = ""
      out2 = ""
      new_context(out) do
        get_mainstream << "aaa"
        new_context(out2) do
          get_mainstream << "111"
          get_mainstream << "222"
        end
        get_mainstream << "bbb"
      end
      assert_equal("aaabbb", out)
      assert_equal("111222", out2)
    end

    def test_substream
      out = ""
      new_context(out) do
        get_mainstream << "aaa"
        new_context(get_substream(:xxx)) do
          get_mainstream << "111"
          get_mainstream << "222"
        end
        output_substream(:xxx)
        get_mainstream << "bbb"
      end
      assert_equal("aaa111222bbb", out)
    end

    def test_substream2
      out = ""
      new_context(out) do
        new_context(get_substream(:xxx)) do
          get_mainstream << "1"
          get_mainstream << "2"
        end
        new_context(get_substream(:yyy)) do
          get_mainstream << "3"
          get_mainstream << "4"
        end
        get_mainstream << "aaa"
        output_substream(:xxx)
        get_mainstream << "bbb"
        output_substream(:yyy)
        get_mainstream << "ccc"
      end
      assert_equal("aaa12bbb34ccc", out)
    end
  end

  class TestDynamicElement < Test::Unit::TestCase
    TEMPLATE_TEXT = <<-END
    <html>
    <body>
      <h1 id="title">title</h1>
      <p id="text">text</p>
      <table>
        <thead>
          <tr id="h"><th></th><th></th></tr>
        </thead>
        <tbody>
          <span></span>
          <tr id="d"><td id="aaa"></td><td id="bbb"></td></tr>
          <span></span>
        </tbody>
      </table>
      <div id="footer">powerd by <a href="http://www.ruby-lang.org/">ruby</a></div>
    </body>
    </html>
    END

    
    def test_dynamic_element
      spec = Amrita2::define_template_spec do 
        dynamic_element(:title)
      end
      assert(spec.kind_of?(RootSpec))
      assert_equal(:title, spec.children[0].sym)
      assert(spec.children[0].kind_of?(DynamicElementSpec))
      assert_equal(spec, spec.children[0].parent)

      doc = REXML::Document.new TEMPLATE_TEXT
      t = spec.create_template(doc)

      root = doc.root
      e = root.elements["body/h1"]
      assert(root.elements["body/h1"].amrita_spec)
      assert(:title, root.elements["body/h1"].amrita_spec.sym)
      assert(root.elements["body/h1"].amrita_spec.parent)
      assert(root.elements["body/h1"].amrita_spec.parent.kind_of?(RootSpec))
    end

    def test_dynamic_element2
      spec = Amrita2::define_template_spec do 
        dynamic_element(:title)
        dynamic_element(:text)
      end
      doc = REXML::Document.new TEMPLATE_TEXT
      t = spec.create_template(doc)

      root = doc.root
      e = root.elements["body/h1"]
      assert(root.elements["body/h1"].amrita_spec)
      assert(:title, root.elements["body/h1"].amrita_spec.sym)
      assert(root.elements["body/p"].amrita_spec)
      assert(:title, root.elements["body/p"].amrita_spec.sym)
    end

    def test_dynamic_element3
      spec = Amrita2::define_template_spec do 
        dynamic_element(:table, :matcher=>TagMatcher.new(:table)) do
          dynamic_element(:header,  :matcher=>TagMatcher.new(:tr))
          dynamic_element(:detail,  :matcher=>TagMatcher.new(:tr))
        end
      end
      doc = REXML::Document.new TEMPLATE_TEXT
      t = spec.create_template(doc)

      root = doc.root
      e = root.elements["body/table/thead/tr"]
      assert_equal("h", e.attributes["id"])
      assert_equal(:header, e.amrita_spec.sym)
      e = root.elements["body/table/tbody/tr"]
      assert_equal("d", e.attributes["id"])
      assert_equal(:detail, e.amrita_spec.sym)
    end

    def test_dynamic_element4
      spec = Amrita2::define_template_spec do 
        dynamic_element(:table, :matcher=>TagMatcher.new(:table)) do
          dynamic_element(:header,  :matcher=>TagMatcher.new(:tr))
          dynamic_element(:detail,  :matcher=>TagMatcher.new(:tr)) do
            dynamic_element(:aaa,  :matcher=>TagMatcher.new(:td))
            dynamic_element(:bbb,  :matcher=>TagMatcher.new(:td))
          end
        end
      end
      spec = Amrita2::define_template_spec(:direct_define=>true) {
        table(:matcher=>TagMatcher.new(:table)) {
          header(:matcher=>TagMatcher.new(:tr))
          detail(:matcher=>TagMatcher.new(:tr)) {
            aaa(:matcher=>TagMatcher.new(:td))
            bbb
          }
        }
        footer {
          link:matcher=>TagMatcher.new(:a)
        }
      }
      doc = REXML::Document.new TEMPLATE_TEXT
      t = spec.create_template(doc)

      root = doc.root
      e = root.elements["body/table"]
      assert_equal("table", e.name)
      e = root.elements["body/table/thead/tr"]
      assert_equal("h", e.attributes["id"])
      assert_equal(:header, e.amrita_spec.sym)
      e = root.elements["body/table/tbody/tr"]
      assert_equal("d", e.attributes["id"])
      assert_equal(:detail, e.amrita_spec.sym)
      e = root.elements.to_a("body/table/tbody/tr/td")
      assert_equal(:aaa, e[0].amrita_spec.sym)
      assert_equal(:bbb, e[1].amrita_spec.sym)
      e = root.elements["body/div"]
      assert_equal(:footer, e.amrita_spec.sym)
      e = root.elements["body/div/a"]
      assert_equal(:link, e.amrita_spec.sym)
    end
  end

  class TestValidation < Test::Unit::TestCase
    TEMPLATE_TEXT = <<-END
    <html>
    <body>
      <h1 id="title">title</h1>
      <p id="text">text</p>
      <table>
        <thead>
          <tr id="h"><th></th><th></th></tr>
        </thead>
        <tbody>
          <span></span>
          <tr id="d"><td id="aaa"></td><td id="bbb"></td></tr>
          <span></span>
        </tbody>
      </table>
      <div id="footer">powerd by <a href="http://www.ruby-lang.org/">ruby</a></div>
    </body>
    </html>
    END

    
    def test_validation
      spec = Amrita2::define_template_spec do 
        dynamic_element(:not_found)
      end
      doc = REXML::Document.new TEMPLATE_TEXT

      assert_raise(RuntimeError) do
        t = spec.create_template(doc)
      end
    end

    def test_validation2
      spec = Amrita2::define_template_spec do 
        dynamic_element(:table, :matcher=>TagMatcher.new(:table)) do 
          dynamic_element(:not_found, :matcher=>TagMatcher.new(:zzzz)) 
        end
      end
      doc = REXML::Document.new TEMPLATE_TEXT
      assert_raise(RuntimeError) do
        t = spec.create_template(doc)
      end
    end

    def test_validation3
      spec = Amrita2::define_template_spec do 
        dynamic_element(:table, :matcher=>TagMatcher.new(:table)) do 
          dynamic_element(:d) do
            dynamic_element(:aaa)
            dynamic_element(:bbb)
            dynamic_element(:not_found)
          end
        end
      end
      doc = REXML::Document.new TEMPLATE_TEXT
      assert_raise(RuntimeError) do
        t = spec.create_template(doc)
      end
    end
  end

  class TestTemplateExpand < Test::Unit::TestCase
    include Amrita2::Runtime
    include Amrita2::HtmlTag

    def test_simple
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa)
      end
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('hello amrita2')
      end

      assert_equal("<x id='aaa'>hello amrita2</x>", out)
    end

    def test_template_expand2
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa)
      end
      doc = REXML::Document.new  "<aaa>1<x id='aaa'>2</x>3</aaa>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('hello amrita2')
      end

      assert_equal("<aaa>1<x id='aaa'>hello amrita2</x>3</aaa>", out)

      out = ""
      t.expand_template(out) do |m|
      end

      assert_equal("<aaa>13</aaa>", out)

    end

    def test_template_expand_list
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :delete_id=>true, :delete_tag=>true)
      end
      doc = REXML::Document.new  "<aaa>1<x id='aaa'>2</x>3</aaa>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('hello amrita2')
      end

      #puts out
      assert_equal("<aaa>1hello amrita23</aaa>", out)

      out = ""
      t.expand_template(out) do |m|
        (2..5).each do |i|
          m.aaa(i)
        end
      end

      #puts out
      assert_equal("<aaa>123453</aaa>", out)
    end

    def test_template_expand_list_attr
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :delete_id=>true, :replace_attr=>true)
      end
      doc = REXML::Document.new  "<aaa>1<x id='aaa'>2</x>3</aaa>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('hello amrita2', :a=>'b')
      end

      #puts out
      assert_equal("<aaa>1<x a='b'>hello amrita2</x>3</aaa>", out)

      out = ""
      t.expand_template(out) do |m|
        (2..5).each do |i|
          m.aaa(i, :class=>"cls#{i}")
        end
      end

      #puts out
      assert_equal("<aaa>1<x class='cls2'>2</x><x class='cls3'>3</x><x class='cls4'>4</x><x class='cls5'>5</x>3</aaa>", out)
    end

    def test_template_expand3
      spec = Amrita2::define_template_spec do 
        dynamic_element(:title)
        dynamic_element(:text)
      end
      doc = REXML::Document.new <<-END
      <html>
        <body>
          <h1 id="title">title</h1>
          <p id="text">text</p>
        </body>
      </html>
      END
      t = spec.create_template(doc)

      root = doc.root
      e = root.elements["body/h1"]
      assert(root.elements["body/h1"].amrita_spec)
      assert(:title, root.elements["body/h1"].amrita_spec.sym)

      out = ""
      t.expand_template(out) do |context|
        assert_equal("", get_substream(:title))
        assert_equal("", get_substream(:text))
        context.title('hello amrita2')
        assert_equal("<h1 id='title'>hello amrita2</h1>", get_substream(:title))
        assert_equal("", get_substream(:text))
        context.text('This text was inserted by amrita2')
        assert_equal("<p id='text'>This text was inserted by amrita2</p>", get_substream(:text))
      end

      assert_equal(<<-END.strip, out.strip)
      <html>
        <body>
          <h1 id='title'>hello amrita2</h1>
          <p id='text'>This text was inserted by amrita2</p>
        </body>
      </html>

      END
    end

    def test_template_expand4
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa) do
          dynamic_element(:bbb)
        end
      end
      doc = REXML::Document.new "<a id='aaa'><b id='bbb'>xxx</b></a>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)
      assert(root.elements["b"].amrita_spec)
      assert(:aaa, root.elements["b"].amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa do |m2|
          m2.bbb('yyy')
        end
      end

      assert_equal("<a id='aaa'><b id='bbb'>yyy</b></a>", out)
    end

    def test_expand_table
      spec = Amrita2::define_template_spec(:direct_define=>true) {
        title
        text
        table(:matcher=>TagMatcher.new(:table)) {
          detail(:delete_id=>true) {
            aaa(:matcher=>TagMatcher.new(:td), :delete_id=>true)
            bbb(:delete_id=>true)
          }
        }
        footer {
          link:matcher=>TagMatcher.new(:a)
        }
      }
      doc = REXML::Document.new <<-END
      <html>
        <body>
        <h1 id="title">title</h1>
        <p id="text">text</p>
        <table border='1'>
          <thead>
            <tr><th>aaa</th><th>bbb</th></tr>
          </thead>
          <tbody>
            <tr id="detail"><td id="aaa"></td><td id="bbb"></td></tr>
          </tbody>
        </table>
        <div id="footer">powerd by <a href="http://www.ruby-lang.org/">ruby</a></div>
        </body>
      </html>
      END
      t = spec.create_template(doc)
      out = ""
      t.expand_template(out) do |m|
        m.title("t")
        m.text("tt")
        m.table do |m2|
          3.times do |i|
            m2.detail do |m3|
              m3.aaa('a' + i.to_s)
              m3.bbb('b' + i.to_s)
            end
          end
        end
      end
      assert_equal(<<-END.strip, out.strip)
      <html>
        <body>
        <h1 id='title'>t</h1>
        <p id='text'>tt</p>
        <table border='1'>
          <thead>
            <tr><th>aaa</th><th>bbb</th></tr>
          </thead>
          <tbody>
            <tr><td>a0</td><td>b0</td></tr><tr><td>a1</td><td>b1</td></tr><tr><td>a2</td><td>b2</td></tr>
          </tbody>
        </table>
        
        </body>
      </html>
      END
    end

    def test_list
      doc = REXML::Document.new <<-END
      <body>111<ul>222<li>333:<a href=''>444</a>555</li>666</ul>7777</body>
      END

      spec = Amrita2::define_template_spec(:direct_define=>true) {
        list(:matcher=>UL) {
          item(:matcher=>LI) {
            link(:matcher=>A, :replace_attr=>true)
          }
        }
      }
      t = spec.create_template(doc)
      out = ""
      t.expand_template(out) do |m|
        m.list do |ml|
          ml.item do |mi|
            mi.link('aaa', :href=>'bbb')
          end
        end
      end
      assert_equal("<body>111<ul>222<li>333:<a href='bbb'>aaa</a>555</li>666</ul>7777</body>", out.strip)

      out = ""
      t.expand_template(out) do |m|
        m.list do |ml|
          ml.item do |mi|
            mi.link('a', :href=>'b')
          end
          ml.item do |mi|
            mi.link('x', :href=>'y')
          end
        end
      end
      assert_equal("<body>111<ul>222<li>333:<a href='b'>a</a>555</li><li>333:<a href='y'>x</a>555</li>666</ul>7777</body>",
                   out)
    end

    def test_dl
      spec = Amrita2::define_template_spec do 
        dynamic_element(:l, :matcher=>Amrita2::HtmlTag::DL) do
          dynamic_element(:t, :matcher=>Amrita2::HtmlTag::DT)
          dynamic_element(:d, :matcher=>Amrita2::HtmlTag::DD, :output_stream=>:t)
        end
      end
      doc = REXML::Document.new  "<body><dl><dt /><dd /></dl></body>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.l do |ml|
          ml.t('amrita')
          ml.d('a template library')
          ml.t('Amrita2')
          ml.d('a better template library')
        end
      end
      assert_equal("<body><dl><dt>amrita</dt><dd>a template library</dd><dt>Amrita2</dt><dd>a better template library</dd></dl></body>", out)

    end
  end

  class TestAttr < Test::Unit::TestCase
    def test_replace_attr
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :replace_attr=>true)
      end

      doc = REXML::Document.new  "<x id='aaa' class='c'>111</x>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx')
      end
      assert_equal("<x class='c' id='aaa'>xxx</x>", out)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx', :class=>'cc')
      end
      assert_equal("<x class='cc' id='aaa'>xxx</x>", out)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx', :xxx=>'yyy')
      end

      assert("<x class='c' xxx='yyy' id='aaa'>xxx</x>" ==  out ||
               "<x xxx='yyy' class='c' id='aaa'>xxx</x>" ==  out 
             )


      out = ""
      t.expand_template(out) do |m|
        m.aaa('xxx', :id=>nil, :xxx=>'yyy')
      end
      assert("<x class='c' xxx='yyy'>xxx</x>" == out ||
             "<x xxx='yyy' class='c'>xxx</x>" == out 
             )
    end

    def test_replace_attr_nest
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :replace_attr=>true) do
          dynamic_element(:bbb)
        end
      end
      doc = REXML::Document.new "<a id='aaa' class='abc'><b id='bbb'>xxx</b></a>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa(:class=>'xyz') do |m2|
          m2.bbb('yyy')
        end
      end

      assert_equal("<a class='xyz' id='aaa'><b id='bbb'>yyy</b></a>", out)
    end

    def test_dash
      spec = Amrita2::define_template_spec do 
        dynamic_element(:meta, :replace_attr=>true)
      end

      doc = REXML::Document.new  "<meta id='meta' http-equiv='refresh' content='c' />"
      t = spec.create_template(doc)

      out = ""

      t.expand_template(out) do |m|
        m.meta(nil, :content=>10)
      end
      assert_equal("<meta id='meta' content='10' http-equiv='refresh'></meta>", out)
    end
  end

  class TestPartsTemplate < Test::Unit::TestCase
    include Amrita2

    def test_simple_parts
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa)
      end
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      main = spec.create_template(doc)

      out = ""
      main.expand_template(out) do |m|
        m.aaa('hello amrita2')
      end
      assert_equal("<x id='aaa'>hello amrita2</x>", out)

      parts_spec = Amrita2::define_template_spec do 
        dynamic_element(:xxx) do
          parts_element(:parts1, :delete_id=>true)
          parts_element(:parts2, :delete_id=>true)
        end
      end

      doc = REXML::Document.new  "<zz id='xxx'><a id='parts1' href='xxxx'>111</a><a id='parts2' href='yyyy'>111</a></zz>"
      parts = parts_spec.create_template_module(doc)

      out = ""
      main.expand_template(out) do |m|
        m.aaa(parts::parts1('p1'))
      end
      assert_equal("<x id='aaa'><a href='xxxx'>p1</a></x>", out)

      out = ""
      main.expand_template(out) do |m|
        m.aaa(parts.parts1('p1'))
        m.aaa(parts.parts1('p1'))
      end
      assert_equal("<x id='aaa'><a href='xxxx'>p1</a></x><x id='aaa'><a href='xxxx'>p1</a></x>", out)

      out = ""
      main.expand_template(out) do |m|
        m.aaa(parts::parts1('p1'))
        m.aaa(SanitizedString[parts::parts2('p2')+'b'])
      end

      assert_equal("<x id='aaa'><a href='xxxx'>p1</a></x><x id='aaa'><a href='yyyy'>p2</a>b</x>", out)
    end

    def test_nested
      parts_spec = Amrita2::define_template_spec do 
        dynamic_element(:section, :replace_attr=>true) {
          dynamic_element(:text, :delete_tag=>true)
        }
      end

      doc = REXML::Document.new  "<section id='section'><span id='text'></span></section>"
      parts_module = parts_spec.create_template_module(doc)
      o = Object.new
      o.extend Amrita2::Runtime
      o.extend parts_module::MainContext
      def o.sec(title, &block) 
        new_context(ret="") do
          section(:title=>title, :id=>nil) do |m|
            m.text(SanitizedString[yield])
          end
          output_substream(:section)
          get_mainstream
        end
        ret
      end

      out = o.sec('title 1') { 'body' }
      assert_equal("<section title='title 1'>body</section>", out)

      out = o.instance_eval do
        s1 = sec('title 1') { 'body' }
        s2 = sec('outer') { sec('inner') { 'body' } }
        s3 = sec('1') do
          sec('1.1')  do
            sec('1.1.1') { '1.1.1 body' } +
              sec('1.1.2') { '1.1.2 body' } 
          end + sec('1.2')  do
            sec('1.2.1') { '1.2.1 body' } 
          end + sec('1.3')  do
            sec('1.3.1') { '1.3.1 body' } +
              sec('1.3.2') { '1.3.2 body' } +
              sec('1.3.3') { '1.3.3 body' } 
          end
        end
      end
      assert_equal("<section title='1'><section title='1.1'><section title='1.1.1'>1.1.1 body</section><section title='1.1.2'>1.1.2 body</section></section><section title='1.2'><section title='1.2.1'>1.2.1 body</section></section><section title='1.3'><section title='1.3.1'>1.3.1 body</section><section title='1.3.2'>1.3.2 body</section><section title='1.3.3'>1.3.3 body</section></section></section>", out)
    end
    def test_outputstream
      spec = Amrita2::define_template_spec do 
        dynamic_element(:table,:matcher=>TagMatcher.new(:table)) {
          dynamic_element(:line1, :replace_attr=>true) {
            dynamic_element(:no, :matcher=>TagMatcher.new(:td))
            dynamic_element(:language, :matcher=>TagMatcher.new(:td))
            dynamic_element(:template, :matcher=>TagMatcher.new(:td))
          }
          dynamic_element(:line2, :replace_attr=>true, :output_stream=>:line1) {
            dynamic_element(:no, :matcher=>TagMatcher.new(:td))
            dynamic_element(:language, :matcher=>TagMatcher.new(:td))
            dynamic_element(:template, :matcher=>TagMatcher.new(:td))
          }
        }
      end
      doc = REXML::Document.new <<-END
      <table>
        <tbody>
          <tr id='line1'><td></td><td></td><td></td></tr>
          <tr id='line2'><td></td><td></td><td></td></tr>
        </tbody>
      </table>
      END
      t = spec.create_template(doc)

      data = [
        %w(ruby amrita),
        %w(perl HTMLtemplate),
        %w(python TAL),
        %w(java JSF)
      ]
      out = ""
      t.expand_template(out) do |m|
        m.table do |mt|
          data.each_with_index do |d, i|
            if i%2 == 0
              mt.line1 do |ml|
                ml.no(i+1)
                ml.language(d[0])
                ml.template(d[1])
              end
            else
              mt.line2 do |ml|
                ml.no(i+1)
                ml.language(d[0])
                ml.template(d[1])
              end
            end
          end
        end
      end
      assert_equal("<table>\n        <tbody>\n          <tr id='line1'><td>1</td><td>ruby</td><td>amrita</td></tr><tr id='line2'><td>2</td><td>perl</td><td>HTMLtemplate</td></tr><tr id='line1'><td>3</td><td>python</td><td>TAL</td></tr><tr id='line2'><td>4</td><td>java</td><td>JSF</td></tr>\n          \n        </tbody>\n      </table>", out)
    end
  end

  class TestDummyElement < Test::Unit::TestCase
    include Amrita2::Runtime
    include Amrita2::HtmlTag

    def test_simple
      spec = Amrita2::define_template_spec do 
        dummy_element(:aaa)
      end
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
      end

      assert_equal("", out)
      out = ""
      t.expand_template(out) do |m|
        assert_raise(NoMethodError) { m.aaa(1) }
      end
    end

    def test_dummy
      spec = Amrita2::define_template_spec do 
        dummy_element(:aaa)
        dynamic_element(:bbb)
      end
      doc = REXML::Document.new  "<a><x id='aaa'>111</x><y id='bbb'></y></a>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.bbb('BB')
      end

      assert_equal("<a><y id='bbb'>BB</y></a>", out)
      out = ""
      t.expand_template(out) do |m|
        assert_raise(NoMethodError) { m.aaa(1) }
      end
    end
  end
end

if __FILE__ == $0
  include Amrita2Test
  cname = ARGV.shift
  if cname
    method = ARGV.shift
    if method
      eval "Test::Unit::UI::Console::TestRunner.run(#{cname}.new(\"#{method}\"))"
    else
      cls = eval(cname)
      Test::Unit::UI::Console::TestRunner.run(cls)
    end
  else
    suite = Test::Unit::TestSuite.new
    suite << TestDummyElement.suite
    suite << TestValidation.suite
    suite << TestSanitize.suite
    suite << TestPartsTemplate.suite
    suite << TestAttr.suite
    suite << TestRuntime.suite
    suite << TestSanitize.suite
    suite << TestTemplateExpand.suite
    suite << TestCodeGenerator.suite
    suite << TestElementMatcher.suite
    suite << TestDynamicElement.suite
    Test::Unit::UI::Console::TestRunner.run(suite)
  end
end
