
require 'test/unit'
require 'test/unit/testsuite'
require 'test/unit/ui/console/testrunner'
require 'amrita2/ext'
require 'rexml/document'

# ruby -I~/amrita2/lib ~/amrita2/test/test_ext.rb   |ruby -pe '$_.sub!(/^ */, "") ' 

module Amrita2ExtTest
  include Amrita2
  include Amrita2::Core

  class TestExpandByMember < Test::Unit::TestCase
    include Amrita2::Runtime
    def test_expand_by_member
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :expand_by_member=>true) do
          dynamic_element(:bbb)
        end
      end
      doc = REXML::Document.new "<a id='aaa'><b id='bbb'>xxx</b></a>"
      t = spec.create_template(doc)

      s = Struct.new(:a, :bbb)
      data = s.new
      data.bbb = 'expand by member'

      out = ""
      t.expand_template(out) do |m|
        m.aaa(data)
      end

      assert_equal("<a id='aaa'><b id='bbb'>expand by member</b></a>", out)
    end

    def test_expand_by_member2
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :expand_by_member=>true, :delete_id=>true) do
          dynamic_element(:bbb, :delete_id=>true)
          dynamic_element(:ccc, :delete_id=>true)
        end
      end
      doc = REXML::Document.new "<a id='aaa'><b id='bbb'>xxx</b><c id='ccc'></c></a>"
      t = spec.create_template(doc)

      s = Struct.new(:a, :bbb, :ccc)
      data = s.new
      data.bbb = 'b'
      data.ccc = 'c'

      out = ""
      t.expand_template(out) do |m|
        m.aaa(data)
        data = s.new
        data.bbb = 'bb'
        data.ccc = 'cc'
        m.aaa(data)
      end

      assert_equal("<a><b>b</b><c>c</c></a><a><b>bb</b><c>cc</c></a>", out)
    end

    def test_expand_by_member3
      spec = Amrita2::define_template_spec do 
        dynamic_element(:time, :expand_by_member=>true, :delete_tag=>true) do
          dynamic_element(:hour, :delete_tag=>true)
          dynamic_element(:min, :delete_tag=>true)
          dynamic_element(:sec, :delete_tag=>true)
        end
      end
      doc = REXML::Document.new "<span id='time'><span id='hour'></span>:<span id='min'></span>:<span id='sec'></span></span>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.time(Time.at(0))
        m.time do |mt|
          get_mainstream << "-"
          mt.hour(11)
          mt.min(20)
          mt.sec(30)
        end
      end

      assert_equal("9:0:0-11:20:30", out)
    end

    def test_outputstream
      spec = Amrita2::define_template_spec do 
        dynamic_element(:table,:matcher=>TagMatcher.new(:table)) {
          dynamic_element(:line1, :expand_by_member=>true, :delete_id=>true) {
            dynamic_element(:no, :matcher=>TagMatcher.new(:td))
            dynamic_element(:language, :matcher=>TagMatcher.new(:td))
            dynamic_element(:template, :matcher=>TagMatcher.new(:td))
          }
          dynamic_element(:line2, :expand_by_member=>true, :output_stream=>:line1, :delete_id=>true) {
            dynamic_element(:no, :matcher=>TagMatcher.new(:td))
            dynamic_element(:language, :matcher=>TagMatcher.new(:td))
            dynamic_element(:template, :matcher=>TagMatcher.new(:td))
          }
        }
      end

      doc = REXML::Document.new <<-END
      <table>
        <tbody>
        <tr id='line1' class='odd'><td></td><td></td><td></td></tr>
        <tr id='line2' class='even'><td></td><td></td><td></td></tr>
        </tbody>
      </table>
        END
      t = spec.create_template(doc)

      s1 = Struct.new('Aaa',  :language, :template, :no)
      data = [
        s1.new('ruby', 'amrita2'),
        s1.new('perl', 'HTMLtemplate'),
        s1.new('python', 'TAL'),
        s1.new('java', 'JSF')
      ]
      out = ""
      t.expand_template(out) do |m|
        m.table do |mt|
          data.each_with_index do |d, i|
            d.no = i+1
            if i%2 == 0
              mt.line1(d)
            else
              mt.line2(d)
            end
          end
        end
      end
      assert_equal("<table>\n        <tbody>\n        <tr class='odd'><td>1</td><td>ruby</td><td>amrita2</td></tr><tr class='even'><td>2</td><td>perl</td><td>HTMLtemplate</td></tr><tr class='odd'><td>3</td><td>python</td><td>TAL</td></tr><tr class='even'><td>4</td><td>java</td><td>JSF</td></tr>\n        \n        </tbody>\n      </table>", out)
    end
  end

  class TestArgs < Test::Unit::TestCase
    def test_args
      spec = Amrita2::define_template_spec do 
        dynamic_element(:raa, :use_args=>true, :delete_id=>true)
      end
      doc = REXML::Document.new  "<a id='raa' href='http://raa.ruby-lang.org/list.rhtml?name=$1'>[$2]</a>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.raa('amrita2', 'amrita2 template library')
      end

      assert_equal("<a href='http://raa.ruby-lang.org/list.rhtml?name=amrita2'>[amrita2 template library]</a>", out)
    end

    def test_args2
      spec = Amrita2::define_template_spec do 
        dynamic_element(:raa, :use_args=>true, :delete_id=>true)
      end
      doc = REXML::Document.new  "<a id='raa' href='http://raa.ruby-lang.org/list.rhtml?name=$1'><em>1$22</em><span class='2$33'>4$04</span>6</a>"

      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.raa('7', '8', '9') { '000' }
      end

      assert_equal("<a href='http://raa.ruby-lang.org/list.rhtml?name=7'><em>182</em><span class='293'>40004</span>6</a>", out)
    end

    def test_sanitize
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :use_args=>true, :delete_id=>true)
      end
      doc = REXML::Document.new  "<x id='aaa' zzz='a$1b'>c$0d</x>"
      t = spec.create_template(doc)

      out = ""
      t.expand_template(out) do |m|
        m.aaa('<>$\'"') { '$\'<>"' }
      end

      assert_equal("<x zzz='a&lt;&gt;$&#39;&quot;b'>c$'&lt;&gt;\"d</x>", out)
    end

  end

  class TestUseBlock < Test::Unit::TestCase
    include Amrita2
    include Amrita2::Runtime

    def test_use_block
      spec = Amrita2::define_template_spec do 
        dynamic_element(:link, :use_original_element=>true)
      end
      doc = REXML::Document.new  "<p>See <a id='link' href='http://www.ruby-lang.org/'>Ruby</a> for detail</p>"
      t = spec.create_template(doc)
      
      t.expand_template(out="") do |m|
        m.link('Ruby homepage')
      end
      assert_equal("<p>See <a href='http://www.ruby-lang.org/' id='link'>Ruby homepage</a> for detail</p>", out)

      t.expand_template(out="") do |m|
        m.link do |e|
          e.attributes["href"] += 'ja/'
          e.attributes.delete("id")
          e.text += '(Japanese)'
          e
        end
      end
      assert_equal("<p>See <a href='http://www.ruby-lang.org/ja/'>Ruby(Japanese)</a> for detail</p>", out)
    end

    def test_use_block_nest
      spec = Amrita2::define_template_spec do 
        dynamic_element(:paragraph, :use_original_element=>true) {
          dynamic_element(:link, :use_original_element=>true)
        }
      end
      doc = REXML::Document.new  "<p id='paragraph' class='a'>See <a id='link' href='http://www.ruby-lang.org/'>Ruby</a> for detail</p>"
      t = spec.create_template(doc)
      
      t.expand_template(out="") do |m|
        m.paragraph do |m|
          m.paragraph do |e|
            e.attributes["class"] = 'b'
            e.attributes.delete("id")
            e
          end
          m.link do |e|
            e.attributes["href"] += 'ja/'
            e.attributes.delete("id")
            e.text += '(Japanese)'
            e
          end
        end
      end
      assert_equal("<p class='b'>See <a href='http://www.ruby-lang.org/ja/'>Ruby(Japanese)</a> for detail</p>", out)
    end
  end

  class TestCheckTrue < Test::Unit::TestCase
    def test_check_true
      spec = Amrita2::define_template_spec do 
        dynamic_element(:aaa, :check_true=>true)
      end
      doc = REXML::Document.new  "<x id='aaa'>111</x>"
      t = spec.create_template(doc)

      root = doc.root
      assert(root.amrita_spec)
      assert(:aaa, root.amrita_spec.sym)

      out = ""
      t.expand_template(out) do |m|
        m.aaa(true)
      end

      assert_equal("<x id='aaa'>111</x>", out)
    end
  end
end

if __FILE__ == $0
  include Amrita2ExtTest
  cname = ARGV.shift
  if cname
    method = ARGV.shift
    if method
      eval "Test::Unit::UI::Console::TestRunner.run(#{cname}.new(\"#{method}\"))"
    else
      cls = eval(cname)
      Test::Unit::UI::Console::TestRunner.run(cls)
    end
  else
    suite = Test::Unit::TestSuite.new
    suite << TestArgs.suite
    suite << TestExpandByMember.suite
    suite << TestUseBlock.suite
    suite << TestCheckTrue.suite
    Test::Unit::UI::Console::TestRunner.run(suite)
  end
end
